package meerkat.bulletinboard.sqlserver;

import java.sql.*;
import java.util.*;

import com.google.protobuf.*;

import meerkat.bulletinboard.*;
import meerkat.bulletinboard.sqlserver.mappers.*;
import static meerkat.bulletinboard.BulletinBoardConstants.*;

import meerkat.comm.CommunicationException;

import meerkat.comm.MessageOutputStream;
import meerkat.crypto.concrete.ECDSASignature;
import meerkat.crypto.concrete.SHA256Digest;

import meerkat.protobuf.BulletinBoardAPI.*;
import meerkat.protobuf.Crypto.Signature;
import meerkat.protobuf.Crypto.SignatureVerificationKey;


import static meerkat.bulletinboard.sqlserver.BulletinBoardSQLServer.SQLQueryProvider.*;

import javax.sql.DataSource;

import meerkat.util.BulletinBoardUtils;
import meerkat.util.TimestampComparator;
import org.springframework.jdbc.core.namedparam.MapSqlParameterSource;
import org.springframework.jdbc.core.namedparam.NamedParameterJdbcTemplate;
import org.springframework.jdbc.support.GeneratedKeyHolder;
import org.springframework.jdbc.support.KeyHolder;



/**
 * This is a generic SQL implementation of the BulletinBoardServer API.
 */
public class BulletinBoardSQLServer implements BulletinBoardServer{

	/**
	 * This interface provides the required implementation-specific data to enable an access to an actual SQL server.
	 * It accounts for differences in languages between SQL DB types and for the different addresses needed to access them.
	 */
	public interface SQLQueryProvider {

		/**
		 * Allowed query types.
		 * Note that each query returned has to comply with the parameter names specified ny getParamNames
		 */
		public static enum QueryType {

			FIND_MSG_ID(
					new String[] {"MsgId"},
					new int[] {Types.BLOB}
			),

			FIND_TAG_ID(
					new String[] {"Tag"},
					new int[] {Types.VARCHAR}
			),

			INSERT_MSG(
					new String[] {"MsgId","TimeStamp","Msg"},
					new int[] {Types.BLOB, Types.TIMESTAMP, Types.BLOB}
			),

			INSERT_NEW_TAG(
					new String[] {"Tag"},
					new int[] {Types.VARCHAR}
			),

			CONNECT_TAG(
					new String[] {"EntryNum","Tag"},
					new int[] {Types.INTEGER, Types.VARCHAR}
			),

			ADD_SIGNATURE(
					new String[] {"EntryNum","SignerId","Signature"},
					new int[] {Types.INTEGER, Types.BLOB, Types.BLOB}
			),

			GET_SIGNATURES(
					new String[] {"EntryNum"},
					new int[] {Types.INTEGER}
			),

			GET_MESSAGES(
					new String[] {},
					new int[] {}
			),

			COUNT_MESSAGES(
					new String[] {},
					new int[] {}
			),

			GET_MESSAGE_STUBS(
					new String[] {},
					new int[] {}
			),

			GET_LAST_MESSAGE_ENTRY(
					new String[] {},
					new int[] {}
			),

			GET_BATCH_MESSAGE_ENTRY(
					new String[] {"SignerId", "BatchId"},
					new int[] {Types.BLOB, Types.INTEGER}
			),

			CHECK_BATCH_LENGTH(
					new String[] {"SignerId", "BatchId"},
					new int[] {Types.BLOB, Types.INTEGER}
			),

			GET_BATCH_MESSAGE_DATA(
					new String[] {"SignerId", "BatchId", "StartPosition"},
					new int[] {Types.BLOB, Types.INTEGER, Types.INTEGER}
			),

			INSERT_BATCH_DATA(
					new String[] {"SignerId", "BatchId", "SerialNum", "Data"},
					new int[] {Types.BLOB, Types.INTEGER, Types.INTEGER, Types.BLOB}
			),

			CONNECT_BATCH_TAG(
					new String[] {"SignerId", "BatchId", "Tag"},
					new int[] {Types.BLOB, Types.INTEGER, Types.VARCHAR}
			),

			GET_BATCH_TAGS(
					new String[] {"SignerId", "BatchId"},
					new int[] {Types.BLOB, Types.INTEGER}
			),

			REMOVE_BATCH_TAGS(
					new String[] {"SignerId", "BatchId"},
					new int[] {Types.BLOB, Types.INTEGER}
			);

			private String[] paramNames;
			private int[] paramTypes;

			private QueryType(String[] paramNames, int[] paramTypes) {
				this.paramNames = paramNames;
				this.paramTypes = paramTypes;
			}

			public String[] getParamNames() {
				return paramNames;
			}

			public String getParamName(int num) {
				return paramNames[num];
			}

			public int[] getParamTypes() {
				return paramTypes;
			}

			public int getParamType(int num) {
				return paramTypes[num];
			}

		}

		/**
		 * This enum provides the standard translation between a filter type and the corresponding parameter name in the SQL query
		 */
		public static enum FilterTypeParam {

			ENTRY_NUM("EntryNum", Types.INTEGER),
			MSG_ID("MsgId", Types.BLOB),
			SIGNER_ID("SignerId", Types.BLOB),
			TAG("Tag", Types.VARCHAR),
			LIMIT("Limit", Types.INTEGER),
			TIMESTAMP("TimeStamp", Types.TIMESTAMP);

			private FilterTypeParam(String paramName, int paramType) {
				this.paramName = paramName;
				this.paramType = paramType;
			}

			private String paramName;
			private int paramType;

			public static FilterTypeParam getFilterTypeParamName(FilterType filterType) {
				switch (filterType) {

					case MSG_ID:
						return MSG_ID;

					case EXACT_ENTRY:		// Go through
					case MAX_ENTRY:			// Go through
					case MIN_ENTRY:
						return ENTRY_NUM;

					case SIGNER_ID:
						return SIGNER_ID;

					case TAG:
						return TAG;

					case MAX_MESSAGES:
						return LIMIT;

					case BEFORE_TIME:		// Go through
					case AFTER_TIME:
						return TIMESTAMP;

					default:
						return null;
				}
			}

			public String getParamName() {
				return paramName;
			}

			public int getParamType() {
				return paramType;
			}
		}

		/**
		 * This function translates a QueryType into an actual SQL query.
		 * @param queryType is the type of query requested
         * @return a string representation of the query for the specific type of SQL database implemented.
         */
		public String getSQLString(QueryType queryType) throws IllegalArgumentException;

		/**
		 * Used to retrieve a condition to add to an SQL statement that will make the result comply with the filter type
		 * @param filterType is the filter type
		 * @param serialNum is a unique number used to identify the condition variables from other condition instances
		 * @return The SQL string for the condition
		 * @throws IllegalArgumentException if the filter type used is not supported
		 */
		public String getCondition(FilterType filterType, int serialNum) throws IllegalArgumentException;

		public String getConditionParamTypeName(FilterType filterType) throws IllegalArgumentException;

		/**
		 * @return the string needed in order to connect to the DB.
         */
		public DataSource getDataSource();

		/**
		 * This is used to get a list of queries that together create the schema needed for the DB.
		 * Note that these queries should not assume anything about the current state of the DB.
		 * In particular: they should not erase any existing tables and/or entries.
		 * @return The list of queries.
         */
		public List<String> getSchemaCreationCommands();

		/**
		 * This is used to get a list of queries that together delete the schema needed for the DB.
		 * This is useful primarily for tests, in which we want to make sure we start with a clean DB.
		 * @return The list of queries.
		 */
		public List<String> getSchemaDeletionCommands();

	}

	/**
	 * This method returns the value of the parameter specified in a message filter
	 * @param messageFilter is the filter
	 * @return the object parameter for the SQL query embedded in the filter (this depends on the filter type)
     */
	private Object getParam(MessageFilter messageFilter) {

		switch (messageFilter.getType()) {

			case MSG_ID:		// Go through
			case SIGNER_ID:
				return messageFilter.getId().toByteArray();

			case EXACT_ENTRY:	// Go through
			case MAX_ENTRY:		// Go through
			case MIN_ENTRY:
				return messageFilter.getEntry();

			case TAG:
				return messageFilter.getTag();

			case MAX_MESSAGES:
				return messageFilter.getMaxMessages();

			case BEFORE_TIME:	// Go through
			case AFTER_TIME:
				return BulletinBoardUtils.toSQLTimestamp(messageFilter.getTimestamp());

			default:	// Unsupported filter type
				return null;
		}

	}

	/**
	 * This class implements a comparator for the MessageFilter class
	 * The comparison is done solely by comparing the type of the filter
	 * This is used to sort the filters by type
	 */
	public class FilterTypeComparator implements Comparator<MessageFilter> {

		@Override
		public int compare(MessageFilter filter1, MessageFilter filter2) {
			return filter1.getTypeValue() - filter2.getTypeValue();
		}
	}

	protected SQLQueryProvider sqlQueryProvider;

	protected NamedParameterJdbcTemplate jdbcTemplate;
	
	protected BatchDigest digest;
	protected BatchDigitalSignature signer;
	
	protected List<SignatureVerificationKey> trusteeSignatureVerificationArray;
	protected int minTrusteeSignatures;

	protected List<List<SignatureVerificationKey>> pollingCommitteeSignatureVerificationKeyArray;
	protected int minCommiteeSignatures;

	/**
	 * This constructor sets the type of SQL language in use.
	 * @param sqlQueryProvider is the provider of the SQL query strings required for actual operation of the server.
     */
	public BulletinBoardSQLServer(SQLQueryProvider sqlQueryProvider) {
		this.sqlQueryProvider = sqlQueryProvider;
	}

	/**
	 * This method creates the schema in the given DB to prepare for future transactions
	 * It does not assume anything about the current state of the database
	 * @throws SQLException
     */
	private void createSchema() throws SQLException {

		for (String command : sqlQueryProvider.getSchemaCreationCommands()) {
			jdbcTemplate.update(command,(Map) null);
		}

	}

	/**
	 * This method initializes the signatures, connects to the DB and creates the schema (if required).
	 */
	@Override
	public void init(String meerkatDB) throws CommunicationException {
		// TODO write signature reading part.
		
		digest = new GenericBatchDigest(new SHA256Digest());
		signer = new GenericBatchDigitalSignature(new ECDSASignature());

		jdbcTemplate = new NamedParameterJdbcTemplate(sqlQueryProvider.getDataSource());

		try {
			createSchema();
		} catch (SQLException e) {
			throw new CommunicationException("Couldn't create schema " + e.getMessage());
		}

	}

	/**
	 * This method verifies the authenticity of the received message based on
	 * the stored signatures.
	 * 
	 * @param msg
	 *            is the message to authenticate (containing the signature).
	 * @return TRUE if the message is authenticated and FALSE otherwise.
	 */
	private boolean verifyMessage(BulletinBoardMessage msg) {
		//TODO: Replace with actual verification.
		return true;
	}
	
	/**
	 * This procedure makes sure that all tags in the given list have an entry in the tags table. 
	 * @param tags
	 */
	protected void insertNewTags(String[] tags) throws SQLException {

		String sql;

		sql = sqlQueryProvider.getSQLString(QueryType.INSERT_NEW_TAG);
		Map namedParameters[] = new HashMap[tags.length];

		for (int i = 0 ; i < tags.length ; i++){
			namedParameters[i] = new HashMap();
			namedParameters[i].put(QueryType.INSERT_NEW_TAG.getParamName(0), tags[i]);
		}

		jdbcTemplate.batchUpdate(sql, namedParameters);

	}
	
	/**
	 * This procedure is used to convert a boolean to a BoolMsg.
	 * @param b is the boolean to convert.
	 * @return a ProtoBuf message with boolean payload.
	 */
	private BoolMsg boolToBoolMsg(boolean b){
		return BoolMsg.newBuilder()
			.setValue(b)
			.build();
	}


	/**
	 * This method posts a messages to the server
	 * @param msg is the message to post
	 * @param checkSignature decides whether ot not the method should check the signature before it posts the message
	 * @return TRUE if the post is successful and FALSE otherwise
	 * @throws CommunicationException
     */
	public BoolMsg postMessage(BulletinBoardMessage msg, boolean checkSignature) throws CommunicationException{

		if (checkSignature && !verifyMessage(msg)) {
			return boolToBoolMsg(false);
		}

		String sql;
		Map[] namedParameterArray;

		byte[] msgID;
		long entryNum;

		ProtocolStringList tagList;
		String[] tags;

		List<Signature> signatureList;
		Signature[] signatures;

		// Calculate message ID (depending only on the the unsigned message)

		digest.reset();
		digest.update(msg.getMsg());

		msgID = digest.digest();

		// Add message to table if needed and store entry number of message.

		sql = sqlQueryProvider.getSQLString(QueryType.FIND_MSG_ID);
		Map namedParameters = new HashMap();

		namedParameters.put(QueryType.FIND_MSG_ID.getParamName(0),msgID);

		List<Long> entryNums = jdbcTemplate.query(sql, namedParameters, new LongMapper());

		if (entryNums.size() > 0){

			entryNum = entryNums.get(0);

		} else{

			sql = sqlQueryProvider.getSQLString(QueryType.INSERT_MSG);

			namedParameters.put(QueryType.INSERT_MSG.getParamName(1), BulletinBoardUtils.toSQLTimestamp(msg.getMsg().getTimestamp()));
			namedParameters.put(QueryType.INSERT_MSG.getParamName(2), msg.getMsg().toByteArray());

			KeyHolder keyHolder = new GeneratedKeyHolder();
			jdbcTemplate.update(sql, new MapSqlParameterSource(namedParameters), keyHolder);

			entryNum = keyHolder.getKey().longValue();

		}

		// Retrieve tags and store new ones in tag table.

		try {

			tagList = msg.getMsg().getTagList();
			tags = new String[tagList.size()];
			tags = tagList.toArray(tags);

			insertNewTags(tags);

		} catch (SQLException e) {
			throw new CommunicationException(e.getMessage());
		}

		// Connect message to tags.

		sql = sqlQueryProvider.getSQLString(QueryType.CONNECT_TAG);

		namedParameterArray = new HashMap[tags.length];

		for (int i = 0 ; i < tags.length ; i++) {
			namedParameterArray[i] = new HashMap();
			namedParameterArray[i].put(QueryType.CONNECT_TAG.getParamName(0), entryNum);
			namedParameterArray[i].put(QueryType.CONNECT_TAG.getParamName(1), tags[i]);
		}

		jdbcTemplate.batchUpdate(sql, namedParameterArray);

		// Retrieve signatures.

		signatureList = msg.getSigList();
		signatures = new Signature[signatureList.size()];
		signatures = signatureList.toArray(signatures);

		// Connect message to signatures.

		sql = sqlQueryProvider.getSQLString(QueryType.ADD_SIGNATURE);

		namedParameterArray = new HashMap[signatures.length];

		for (int i = 0 ; i < signatures.length ; i++) {
			namedParameterArray[i] = new HashMap();
			namedParameterArray[i].put(QueryType.ADD_SIGNATURE.getParamName(0), entryNum);
			namedParameterArray[i].put(QueryType.ADD_SIGNATURE.getParamName(1), signatures[i].getSignerId().toByteArray());
			namedParameterArray[i].put(QueryType.ADD_SIGNATURE.getParamName(2), signatures[i].toByteArray());
		}

		jdbcTemplate.batchUpdate(sql,namedParameterArray);

		return boolToBoolMsg(true);

	}

	@Override
	public BoolMsg postMessage(BulletinBoardMessage msg) throws CommunicationException {
		return postMessage(msg, true);	// Perform a post and check the signature for authenticity
	}


	/**
	 * This is a container class for and SQL string builder and a MapSqlParameterSource to be used with it
	 */
	class SQLAndParameters {

		public StringBuilder sql;
		public MapSqlParameterSource parameters;

		public SQLAndParameters(int numOfFilters) {
			sql = new StringBuilder(50 * numOfFilters);
			parameters = new MapSqlParameterSource();
		}

	}

	SQLAndParameters getSQLFromFilters(MessageFilterList filterList) {

		SQLAndParameters result = new SQLAndParameters(filterList.getFilterCount());

		List<MessageFilter> filters = new ArrayList<MessageFilter>(filterList.getFilterList());

		Collections.sort(filters, new FilterTypeComparator());

		boolean isFirstFilter = true;

		if (!filters.isEmpty()) {
			result.sql.append(" WHERE ");

			for (int paramNum = 0 ; paramNum < filters.size() ; paramNum++) {

				MessageFilter filter = filters.get(paramNum);

				if (filter.getType().getNumber() != FilterType.MAX_MESSAGES_VALUE) {
					if (isFirstFilter) {
						isFirstFilter = false;
					} else {
						result.sql.append(" AND ");
					}
				}

				result.sql.append(sqlQueryProvider.getCondition(filter.getType(), paramNum));

				FilterTypeParam filterTypeParam = FilterTypeParam.getFilterTypeParamName(filter.getType());

				result.parameters.addValue(
						filterTypeParam.getParamName() + Integer.toString(paramNum),
						getParam(filter),
						filterTypeParam.getParamType(),
						sqlQueryProvider.getConditionParamTypeName(filter.getType()));

			}

		}

		return result;

	}


	/**
	 * Used to retrieve just basic information about messages to allow calculation of checksum
	 * @param filterList is a filter list that defines which messages the client is interested in
	 * @return a list of Bulletin Board Messages that contain just the entry number, timestamp and message ID for each message
	 * 			The message ID is returned inside the message data field
     */
	protected List<BulletinBoardMessage> readMessageStubs(MessageFilterList filterList) {

		StringBuilder sqlBuilder = new StringBuilder(50 * (filterList.getFilterCount() + 1));

		sqlBuilder.append(sqlQueryProvider.getSQLString(QueryType.GET_MESSAGE_STUBS));

		// Get Conditions

		SQLAndParameters sqlAndParameters = getSQLFromFilters(filterList);

		sqlBuilder.append(sqlAndParameters.sql);

		// Run query

		return jdbcTemplate.query(sqlBuilder.toString(), sqlAndParameters.parameters, new MessageStubMapper());

	}


	@Override
	public void readMessages(MessageFilterList filterList, MessageOutputStream<BulletinBoardMessage> out) throws CommunicationException {

		BulletinBoardMessageList.Builder resultListBuilder = BulletinBoardMessageList.newBuilder();

		// SQL length is roughly 50 characters per filter + 50 for the query itself
		StringBuilder sqlBuilder = new StringBuilder(50 * (filterList.getFilterCount() + 1));

		// Check if Tag/Signature tables are required for filtering purposes

		sqlBuilder.append(sqlQueryProvider.getSQLString(QueryType.GET_MESSAGES));

		// Get conditions

		SQLAndParameters sqlAndParameters = getSQLFromFilters(filterList);
		sqlBuilder.append(sqlAndParameters.sql);

		// Run query and stream the output using a MessageCallbackHandler

		jdbcTemplate.query(sqlBuilder.toString(), sqlAndParameters.parameters, new MessageCallbackHandler(jdbcTemplate, sqlQueryProvider, out));

	}

	/**
	 * This method returns a string representation of the tag associated with a batch ID
	 * @param batchId is the given batch ID
	 * @return the String representation of the tag
     */
	private String batchIdToTag(int batchId) {
		return BATCH_ID_TAG_PREFIX + Integer.toString(batchId);
	}

	/**
	 * This method checks if a specified batch exists and is already closed
	 * @param signerId is the ID of the publisher of the batch
	 * @param batchId is the unique (per signer) batch ID
     * @return TRUE if the batch is closed and FALSE if it is still open or doesn't exist at all
     */
	private boolean isBatchClosed(ByteString signerId, int batchId) throws CommunicationException {

		MessageFilterList filterList = MessageFilterList.newBuilder()
				.addFilter(MessageFilter.newBuilder()
						.setType(FilterType.SIGNER_ID)
						.setId(signerId)
						.build())
				.addFilter(MessageFilter.newBuilder()
						.setType(FilterType.TAG)
						.setTag(BATCH_TAG)
						.build())
				.addFilter(MessageFilter.newBuilder()
						.setType(FilterType.TAG)
						.setTag(batchIdToTag(batchId))
						.build())
				.build();

		// SQL length is roughly 50 characters per filter + 50 for the query itself
		StringBuilder sqlBuilder = new StringBuilder(50 * (filterList.getFilterCount() + 1));

		// Check if Tag/Signature tables are required for filtering purposes

		sqlBuilder.append(sqlQueryProvider.getSQLString(QueryType.COUNT_MESSAGES));

		// Get conditions

		SQLAndParameters sqlAndParameters = getSQLFromFilters(filterList);
		sqlBuilder.append(sqlAndParameters.sql);

		// Run query and stream the output using a MessageCallbackHandler

		List<Long> count = jdbcTemplate.query(sqlBuilder.toString(), sqlAndParameters.parameters, new LongMapper());

		return (count.size() > 0) && (count.get(0) > 0);

	}

	@Override
	public BoolMsg beginBatch(BeginBatchMessage message) throws CommunicationException {

		// Check if batch is closed
		if (isBatchClosed(message.getSignerId(), message.getBatchId())) {
			return BoolMsg.newBuilder().setValue(false).build();
		}

		// Add new tags to table
		ProtocolStringList tagList = message.getTagList();
		String[] tags = new String[tagList.size()];
		tags = tagList.toArray(tags);
		try {
			insertNewTags(tags);
		} catch (SQLException e) {
			throw new CommunicationException(e.getMessage());
		}

		// Connect tags
		String sql = sqlQueryProvider.getSQLString(QueryType.CONNECT_BATCH_TAG);
		MapSqlParameterSource namedParameters[] = new MapSqlParameterSource[tags.length];

		for (int i=0 ; i < tags.length ; i++) {
			namedParameters[i] = new MapSqlParameterSource();
			namedParameters[i].addValue(QueryType.CONNECT_BATCH_TAG.getParamName(0),message.getSignerId().toByteArray());
			namedParameters[i].addValue(QueryType.CONNECT_BATCH_TAG.getParamName(1),message.getBatchId());
			namedParameters[i].addValue(QueryType.CONNECT_BATCH_TAG.getParamName(2),tags[i]);
		}

		jdbcTemplate.batchUpdate(sql,namedParameters);

		return BoolMsg.newBuilder().setValue(true).build();
	}

	@Override
	public BoolMsg postBatchMessage(BatchMessage batchMessage) throws CommunicationException{

		// Check if batch is closed
		if (isBatchClosed(batchMessage.getSignerId(), batchMessage.getBatchId())) {
			return BoolMsg.newBuilder().setValue(false).build();
		}

		// Add data
		String sql = sqlQueryProvider.getSQLString(QueryType.INSERT_BATCH_DATA);
		MapSqlParameterSource namedParameters = new MapSqlParameterSource();

		namedParameters.addValue(QueryType.INSERT_BATCH_DATA.getParamName(0),batchMessage.getSignerId().toByteArray());
		namedParameters.addValue(QueryType.INSERT_BATCH_DATA.getParamName(1),batchMessage.getBatchId());
		namedParameters.addValue(QueryType.INSERT_BATCH_DATA.getParamName(2),batchMessage.getSerialNum());
		namedParameters.addValue(QueryType.INSERT_BATCH_DATA.getParamName(3),batchMessage.getData().toByteArray());

		jdbcTemplate.update(sql, namedParameters);

		return BoolMsg.newBuilder().setValue(true).build();

	}

	@Override
	public BoolMsg closeBatchMessage(CloseBatchMessage message) throws CommunicationException {

		ByteString signerId = message.getSig().getSignerId();
		int batchId = message.getBatchId();

		KeyHolder keyHolder = new GeneratedKeyHolder();

		// Check batch size

		String sql = sqlQueryProvider.getSQLString(QueryType.CHECK_BATCH_LENGTH);
		MapSqlParameterSource namedParameters = new MapSqlParameterSource();


		namedParameters.addValue(QueryType.CHECK_BATCH_LENGTH.getParamName(0),signerId.toByteArray());
		namedParameters.addValue(QueryType.CHECK_BATCH_LENGTH.getParamName(1),batchId);

		List<Long> lengthResult = jdbcTemplate.query(sql, namedParameters, new LongMapper());

		if (lengthResult.get(0) != message.getBatchLength()) {
			return BoolMsg.newBuilder().setValue(false).build();
		}


		// Get Tags and add them to CompleteBatch

		sql = sqlQueryProvider.getSQLString(QueryType.GET_BATCH_TAGS);
		namedParameters = new MapSqlParameterSource();

		namedParameters.addValue(QueryType.GET_BATCH_TAGS.getParamName(0),signerId);
		namedParameters.addValue(QueryType.GET_BATCH_TAGS.getParamName(1),batchId);

		List<String> tags = jdbcTemplate.query(sql, namedParameters, new StringMapper());

		CompleteBatch completeBatch = new CompleteBatch(
				BeginBatchMessage.newBuilder()
						.setSignerId(signerId)
						.setBatchId(batchId)
						.addAllTag(tags)
						.build()
		);

		// Add timestamp to CompleteBatch
		completeBatch.setTimestamp(message.getTimestamp());

		// Add actual batch data to CompleteBatch

		sql = sqlQueryProvider.getSQLString(QueryType.GET_BATCH_MESSAGE_DATA);
		namedParameters = new MapSqlParameterSource();

		namedParameters.addValue(QueryType.GET_BATCH_MESSAGE_DATA.getParamName(0),signerId.toByteArray());
		namedParameters.addValue(QueryType.GET_BATCH_MESSAGE_DATA.getParamName(1),batchId);
		namedParameters.addValue(QueryType.GET_BATCH_MESSAGE_DATA.getParamName(2),0); // Read from the beginning

		completeBatch.appendBatchData(jdbcTemplate.query(sql, namedParameters, new BatchDataMapper()));

		// Verify signature

		completeBatch.setSignature(message.getSig());

//		try {
//			TODO: Actual verification
//			//signer.verify(completeBatch);
//		} catch (CertificateException | InvalidKeyException | SignatureException e) {
//			return BoolMsg.newBuilder().setValue(false).build();
//		}

		// Batch verified: finalize it

		// Calculate message ID
		digest.reset();
		digest.update(completeBatch);
		MessageID msgID = MessageID.newBuilder().setID(ByteString.copyFrom(digest.digest())).build();

		// Create Bulletin Board message
		BulletinBoardMessage bulletinBoardMessage = BulletinBoardMessage.newBuilder()
				.addSig(message.getSig())
				.setMsg(UnsignedBulletinBoardMessage.newBuilder()
					.addAllTag(tags)
					.addTag(BATCH_TAG)
					.addTag(batchIdToTag(batchId))
					.setData(message.getSig().getSignerId())
					.setTimestamp(message.getTimestamp())
					.build())
				.build();

		// Post message without checking signature validity
		postMessage(bulletinBoardMessage, false);

		// Remove tags from temporary table
		sql = sqlQueryProvider.getSQLString(QueryType.REMOVE_BATCH_TAGS);
		namedParameters = new MapSqlParameterSource();

		namedParameters.addValue(QueryType.REMOVE_BATCH_TAGS.getParamName(0), signerId.toByteArray());
		namedParameters.addValue(QueryType.REMOVE_BATCH_TAGS.getParamName(1), batchId);

		jdbcTemplate.update(sql, namedParameters);

		// Return TRUE

		return BoolMsg.newBuilder().setValue(true).build();
	}

	@Override
	public void readBatch(BatchSpecificationMessage message, MessageOutputStream<BatchData> out) throws CommunicationException, IllegalArgumentException{

		// Check that batch is closed
		if (!isBatchClosed(message.getSignerId(), message.getBatchId())) {
			throw new IllegalArgumentException("No such batch");
		}

		String sql = sqlQueryProvider.getSQLString(QueryType.GET_BATCH_MESSAGE_DATA);
		MapSqlParameterSource namedParameters = new MapSqlParameterSource();

		namedParameters.addValue(QueryType.GET_BATCH_MESSAGE_DATA.getParamName(0),message.getSignerId().toByteArray());
		namedParameters.addValue(QueryType.GET_BATCH_MESSAGE_DATA.getParamName(1),message.getBatchId());
		namedParameters.addValue(QueryType.GET_BATCH_MESSAGE_DATA.getParamName(2),message.getStartPosition());

		jdbcTemplate.query(sql, namedParameters, new BatchDataCallbackHandler(out));

	}

	/**
	 * Finds the entry number of the last entry in the database
	 * @return the entry number, or -1 if no entries are found
     */
	protected long getLastMessageEntry() {

		String sql = sqlQueryProvider.getSQLString(QueryType.GET_LAST_MESSAGE_ENTRY);

		List<Long> resultList = jdbcTemplate.query(sql, new LongMapper());

		if (resultList.size() <= 0){
			return -1;
		}

		return resultList.get(0);

	}

	/**
	 * Searches for the latest time of sync of the DB relative to a given query and returns the metadata needed to complete the sync
	 * The checksum up to (and including) each given timestamp is calculated using bitwise XOR on 8-byte sized blocks of the message IDs
	 * @param syncQuery contains a succinct representation of states to compare to
	 * @return the current last entry num and latest time of sync if there is one; -1 as last entry and empty timestamp otherwise
	 * @throws CommunicationException
     */
	@Override
	public SyncQueryResponse querySync(SyncQuery syncQuery) throws CommunicationException {

		if (syncQuery == null){
			return SyncQueryResponse.newBuilder()
					.setLastEntryNum(-1)
					.setLastTimeOfSync(com.google.protobuf.Timestamp.getDefaultInstance())
					.build();
		}

		com.google.protobuf.Timestamp lastTimeOfSync = null;

		TimestampComparator timestampComparator = new TimestampComparator();

		long lastEntryNum = getLastMessageEntry();

		Iterator<SingleSyncQuery> queryIterator = syncQuery.getQueryList().iterator();

		SingleSyncQuery currentQuery = queryIterator.next();

		List<BulletinBoardMessage> messageStubs = readMessageStubs(syncQuery.getFilterList());

		Checksum checksum = new SimpleChecksum();

		for (BulletinBoardMessage message : messageStubs){

			// Check for end of current query
			if (timestampComparator.compare(message.getMsg().getTimestamp(), currentQuery.getTimeOfSync()) > 0){

				if (checksum.getChecksum() == currentQuery.getChecksum()){
					lastTimeOfSync = currentQuery.getTimeOfSync();
				} else {
					break;
				}

				if (queryIterator.hasNext()){
					currentQuery = queryIterator.next();
				} else{
					break;
				}

			}

			// Advance checksum

			ByteString messageID = message.getMsg().getData();	// The data field contains the message ID

			checksum.update(messageID);

		}

		if (checksum.getChecksum() == currentQuery.getChecksum()){
			lastTimeOfSync = currentQuery.getTimeOfSync();
		}

		if (lastTimeOfSync == null){
			return SyncQueryResponse.newBuilder()
					.setLastEntryNum(-1)
					.setLastTimeOfSync(com.google.protobuf.Timestamp.getDefaultInstance())
					.build();
		} else{
			return SyncQueryResponse.newBuilder()
					.setLastEntryNum(lastEntryNum)
					.setLastTimeOfSync(lastTimeOfSync)
					.build();
		}

	}

	@Override
	public void close() {}

}
