package meerkat.bulletinboard.sqlserver;

import com.mysql.jdbc.jdbc2.optional.MysqlDataSource;
import meerkat.bulletinboard.BulletinBoardConstants;
import meerkat.bulletinboard.sqlserver.BulletinBoardSQLServer.SQLQueryProvider;
import meerkat.protobuf.BulletinBoardAPI.FilterType;

import javax.sql.DataSource;
import java.text.MessageFormat;
import java.util.LinkedList;
import java.util.List;

/**
 * Created by Arbel Deutsch Peled on 09-Dec-15.
 */

public class MySQLQueryProvider implements SQLQueryProvider {

    private String dbAddress;
    private int dbPort;
    private String dbName;
    private String username;
    private String password;

    public MySQLQueryProvider(String dbAddress, int dbPort, String dbName, String username, String password) {
        this.dbAddress = dbAddress;
        this.dbPort = dbPort;
        this.dbName = dbName;
        this.username = username;
        this.password = password;
    }

    @Override
    public String getSQLString(QueryType queryType) throws IllegalArgumentException{

        switch(queryType) {

            case ADD_SIGNATURE:
                return MessageFormat.format(
                        "INSERT IGNORE INTO SignatureTable (EntryNum, SignerId, Signature) VALUES (:{0}, :{1}, :{2})",
                        QueryType.ADD_SIGNATURE.getParamName(0),
                        QueryType.ADD_SIGNATURE.getParamName(1),
                        QueryType.ADD_SIGNATURE.getParamName(2));

            case CONNECT_TAG:
                return MessageFormat.format(
                        "INSERT IGNORE INTO MsgTagTable (TagId, EntryNum)"
                        + " SELECT TagTable.TagId, :{0} AS EntryNum FROM TagTable WHERE Tag = :{1}",
                        QueryType.CONNECT_TAG.getParamName(0),
                        QueryType.CONNECT_TAG.getParamName(1));

            case FIND_MSG_ID:
                return MessageFormat.format(
                        "SELECT EntryNum From MsgTable WHERE MsgId = :{0}",
                        QueryType.FIND_MSG_ID.getParamName(0));

            case FIND_TAG_ID:
                return MessageFormat.format(
                        "SELECT TagId FROM TagTable WHERE Tag = :{0}",
                        QueryType.FIND_TAG_ID.getParamName(0));

            case GET_MESSAGES:
                return "SELECT MsgTable.EntryNum, MsgTable.Msg FROM MsgTable";

            case COUNT_MESSAGES:
                return "SELECT COUNT(MsgTable.EntryNum) FROM MsgTable";

            case GET_MESSAGE_STUBS:
                return "SELECT MsgTable.EntryNum, MsgTable.MsgId, MsgTable.ExactTime FROM MsgTable";

            case GET_SIGNATURES:
                return MessageFormat.format(
                        "SELECT Signature FROM SignatureTable WHERE EntryNum = :{0}",
                        QueryType.GET_SIGNATURES.getParamName(0));

            case INSERT_MSG:
                return MessageFormat.format(
                        "INSERT INTO MsgTable (MsgId, ExactTime, Msg) VALUES(:{0}, :{1}, :{2})",
                        QueryType.INSERT_MSG.getParamName(0),
                        QueryType.INSERT_MSG.getParamName(1),
                        QueryType.INSERT_MSG.getParamName(2));

            case INSERT_NEW_TAG:
                return MessageFormat.format(
                        "INSERT IGNORE INTO TagTable(Tag) VALUES (:{0})",
                        QueryType.INSERT_NEW_TAG.getParamName(0));

            case GET_LAST_MESSAGE_ENTRY:
                return "SELECT MAX(MsgTable.EntryNum) FROM MsgTable";

            case GET_BATCH_MESSAGE_ENTRY:
                return MessageFormat.format(
                        "SELECT MsgTable.EntryNum, MsgTable.Msg FROM MsgTable"
                        + " INNER JOIN SignatureTable ON MsgTable.EntryNum = SignatureTable.EntryNum"
                        + " INNER JOIN MsgTagTable ON MsgTable.EntryNum = MsgTagTable.EntryNum"
                        + " INNER JOIN TagTable ON MsgTagTable.TagId = TagTable.TagId"
                        + " WHERE SignatureTable.SignerId = :{0}"
                        + " AND TagTable.Tag = :{1}",
                        QueryType.GET_BATCH_MESSAGE_ENTRY.getParamName(0),
                        QueryType.GET_BATCH_MESSAGE_ENTRY.getParamName(1));

            case GET_BATCH_MESSAGE_DATA:
                return MessageFormat.format(
                        "SELECT Data FROM BatchTable"
                        + " WHERE SignerId = :{0} AND BatchId = :{1} AND SerialNum >= :{2}"
                        + " ORDER BY SerialNum ASC",
                        QueryType.GET_BATCH_MESSAGE_DATA.getParamName(0),
                        QueryType.GET_BATCH_MESSAGE_DATA.getParamName(1),
                        QueryType.GET_BATCH_MESSAGE_DATA.getParamName(2));

            case INSERT_BATCH_DATA:
                return MessageFormat.format(
                        "INSERT INTO BatchTable (SignerId, BatchId, SerialNum, Data)"
                        + " VALUES (:{0}, :{1}, :{2}, :{3})",
                        QueryType.INSERT_BATCH_DATA.getParamName(0),
                        QueryType.INSERT_BATCH_DATA.getParamName(1),
                        QueryType.INSERT_BATCH_DATA.getParamName(2),
                        QueryType.INSERT_BATCH_DATA.getParamName(3));

            case CHECK_BATCH_LENGTH:
                return MessageFormat.format(
                        "SELECT COUNT(Data) AS BatchLength FROM BatchTable"
                        + " WHERE SignerId = :{0} AND BatchId = :{1}",
                        QueryType.CHECK_BATCH_LENGTH.getParamName(0),
                        QueryType.CHECK_BATCH_LENGTH.getParamName(1));

            case CONNECT_BATCH_TAG:
                return MessageFormat.format(
                        "INSERT INTO BatchTagTable (SignerId, BatchId, TagId) SELECT :{0}, :{1}, TagId FROM TagTable"
                        + " WHERE Tag = :{2}",
                        QueryType.CONNECT_BATCH_TAG.getParamName(0),
                        QueryType.CONNECT_BATCH_TAG.getParamName(1),
                        QueryType.CONNECT_BATCH_TAG.getParamName(2));

            case GET_BATCH_TAGS:
                return MessageFormat.format(
                        "SELECT Tag FROM TagTable INNER JOIN BatchTagTable ON TagTable.TagId = BatchTagTable.TagId"
                        + " WHERE SignerId = :{0} AND BatchId = :{1} ORDER BY Tag ASC",
                        QueryType.GET_BATCH_TAGS.getParamName(0),
                        QueryType.GET_BATCH_TAGS.getParamName(1));

            case REMOVE_BATCH_TAGS:
                return MessageFormat.format(
                        "DELETE FROM BatchTagTable WHERE SignerId = :{0} AND BatchId = :{1}",
                        QueryType.REMOVE_BATCH_TAGS.getParamName(0),
                        QueryType.REMOVE_BATCH_TAGS.getParamName(1));

           default:
               throw new IllegalArgumentException("Cannot serve a query of type " + queryType);
        }

    }

    @Override
    public String getCondition(FilterType filterType, int serialNum) throws IllegalArgumentException {

        String serialString = Integer.toString(serialNum);

        switch(filterType) {
            case EXACT_ENTRY:
                return "MsgTable.EntryNum = :EntryNum" + serialString;
            case MAX_ENTRY:
                return "MsgTable.EntryNum <= :EntryNum" + serialString;
            case MIN_ENTRY:
                return "MsgTable.EntryNum >= :EntryNum" + serialString;
            case MAX_MESSAGES:
                return "LIMIT :Limit" + serialString;
            case MSG_ID:
                return "MsgTable.MsgId = :MsgId" + serialString;
            case SIGNER_ID:
                return "EXISTS (SELECT 1 FROM SignatureTable"
                        + " WHERE SignatureTable.SignerId = :SignerId" + serialString + " AND SignatureTable.EntryNum = MsgTable.EntryNum)";
            case TAG:
                return "EXISTS (SELECT 1 FROM TagTable"
                        + " INNER JOIN MsgTagTable ON TagTable.TagId = MsgTagTable.TagId"
                        + " WHERE TagTable.Tag = :Tag" + serialString + " AND MsgTagTable.EntryNum = MsgTable.EntryNum)";

            case BEFORE_TIME:
                return "MsgTable.ExactTime <= :TimeStamp";

            case AFTER_TIME:
                return "MsgTable.ExactTime >= :TimeStamp";

            default:
                throw new IllegalArgumentException("Cannot serve a filter of type " + filterType);
        }

    }

    @Override
    public String getConditionParamTypeName(FilterType filterType) throws IllegalArgumentException {

        switch(filterType) {
            case EXACT_ENTRY:   // Go through
            case MAX_ENTRY:     // Go through
            case MIN_ENTRY:     // Go through
            case MAX_MESSAGES:
                return "INT";

            case MSG_ID:        // Go through
            case SIGNER_ID:
                return "TINYBLOB";

            case TAG:
                return "VARCHAR";

            case AFTER_TIME:    // Go through
            case BEFORE_TIME:
                return "TIMESTAMP";

            default:
                throw new IllegalArgumentException("Cannot serve a filter of type " + filterType);
        }

    }

    @Override
    public DataSource getDataSource() {
        MysqlDataSource dataSource = new MysqlDataSource();

        dataSource.setServerName(dbAddress);
        dataSource.setPort(dbPort);
        dataSource.setDatabaseName(dbName);
        dataSource.setUser(username);
        dataSource.setPassword(password);
        dataSource.setAllowMultiQueries(true);

        return dataSource;
    }

    @Override
    public List<String> getSchemaCreationCommands() {
        List<String> list = new LinkedList<String>();

        list.add("CREATE TABLE IF NOT EXISTS MsgTable (EntryNum INT NOT NULL AUTO_INCREMENT PRIMARY KEY,"
                + " MsgId TINYBLOB, ExactTime TIMESTAMP, Msg BLOB, UNIQUE(MsgId(50)))");

        list.add("CREATE TABLE IF NOT EXISTS TagTable (TagId INT NOT NULL AUTO_INCREMENT PRIMARY KEY, Tag VARCHAR(50), UNIQUE(Tag))");

        list.add("CREATE TABLE IF NOT EXISTS MsgTagTable (EntryNum INT, TagId INT,"
                + " CONSTRAINT FOREIGN KEY (EntryNum) REFERENCES MsgTable(EntryNum),"
                + " CONSTRAINT FOREIGN KEY (TagId) REFERENCES TagTable(TagId),"
                + " CONSTRAINT UNIQUE (EntryNum, TagID))");

        list.add("CREATE TABLE IF NOT EXISTS SignatureTable (EntryNum INT, SignerId TINYBLOB, Signature TINYBLOB,"
                + " INDEX(SignerId(32)), CONSTRAINT Unique_Signature UNIQUE(SignerId(32), EntryNum),"
                + " CONSTRAINT FOREIGN KEY (EntryNum) REFERENCES MsgTable(EntryNum))");

        list.add("CREATE TABLE IF NOT EXISTS BatchTable (SignerId TINYBLOB, BatchId INT, SerialNum INT, Data BLOB,"
                + " CONSTRAINT Unique_Batch UNIQUE(SignerId(32), BatchId, SerialNum))");

        list.add("CREATE TABLE IF NOT EXISTS BatchTagTable (SignerId TINYBLOB, BatchId INT, TagId INT,"
                + " INDEX(SignerId(32), BatchId), CONSTRAINT FOREIGN KEY (TagId) REFERENCES TagTable(TagId))");

        return list;
    }

    @Override
    public List<String> getSchemaDeletionCommands() {
        List<String> list = new LinkedList<String>();

        list.add("DROP TABLE IF EXISTS BatchTagTable");
        list.add("DROP TABLE IF EXISTS BatchTable");
        list.add("DROP TABLE IF EXISTS MsgTagTable");
        list.add("DROP TABLE IF EXISTS SignatureTable");
        list.add("DROP TABLE IF EXISTS TagTable");
        list.add("DROP TABLE IF EXISTS MsgTable");

        return list;
    }
}
